<!-- AUTOGENERATED BY glean_parser v{{ parser_version }}. DO NOT EDIT. -->
{# The rendered markdown is autogenerated, but this
Jinja2 template is not. Please file bugs! #}

# Metrics

This document enumerates the metrics collected by {{ project_title }} using the [Glean SDK](https://mozilla.github.io/glean/book/index.html).
This project may depend on other projects which also collect metrics.
This means you might have to go searching through the dependency tree to get a full picture of everything collected by this project.
{% if introduction_extra %}

{{ introduction_extra }}
{% endif %}

# Pings

{% for ping_name in metrics_by_pings.keys()|sort %}
- [{{ ping_name }}]({{ '#' }}{{ ping_name|replace(" ","-") }})
{% endfor %}

{% for ping_name in metrics_by_pings.keys()|sort %}
{% raw %}##{% endraw %} {{ ping_name }}

{% if ping_name|ping_desc and ping_name|ping_desc|length > 0 %}
{{ ping_name|ping_desc }}

{% if ping_name|ping_docs|length > 0 %}
See the Glean SDK documentation for the [`{{ ping_name }}` ping]({{ ping_name|ping_docs }}).

{% endif %}
{% endif %}
{% if ping_name|ping_send_if_empty %}
This ping is sent if empty.

{% endif %}
{% if ping_name|ping_include_client_id %}
This ping includes the [client id](https://mozilla.github.io/glean/book/user/pings/index.html#the-client_info-section).

{% endif %}
{% if ping_name|ping_data_reviews %}
**Data reviews for this ping:**

{% for review in ping_name|ping_data_reviews %}
- <{{review}}>
{% endfor %}

{% endif %}
{% if ping_name|ping_bugs %}
**Bugs related to this ping:**

{% for bug in ping_name|ping_bugs %}
- {% if bug|int != 0 %}{{bug}}{% else %}<{{bug}}>{% endif %}

{% endfor %}

{% endif %}
{% if ping_name|ping_reasons %}
**Reasons this ping may be sent:**

{% for (reason, desc) in ping_name|ping_reasons|dictsort %}
- `{{ reason }}`: {{ desc|indent(6, first=False) }}
{% endfor %}

{% endif %}
All Glean pings contain built-in metrics in the [`ping_info`](https://mozilla.github.io/glean/book/user/pings/index.html#the-ping_info-section) and [`client_info`](https://mozilla.github.io/glean/book/user/pings/index.html#the-client_info-section) sections.

{% if metrics_by_pings[ping_name] %}
In addition to those built-in metrics, the following metrics are added to the ping:

| Name | Type | Description | Data reviews | Extras | Expiration | [Data Sensitivity](https://wiki.mozilla.org/Firefox/Data_Collection) |
| --- | --- | --- | --- | --- | --- | --- |
{% for metric in metrics_by_pings[ping_name] %}
| {{ metric.identifier() }} |
{{- '['}}{{ metric.type }}]({{ metric.type|metrics_docs }}) |
{{- metric.description|replace("\n", " ") }} |
{%- for data_review in metric.data_reviews %}
[{{ data_review|ping_review_title(loop.index) }}]({{ data_review }}){{ ", " if not loop.last }}
{%- endfor -%} |
{%- if metric|extra_info -%}
<ul>
{%- for property, desc in metric|extra_info %}
<li>{{ property }}{%- if desc is not none -%}: {{ desc|replace("\n", " ") }}{%- endif -%}</li>
{%- endfor -%}
</ul>
{%- endif -%} |
{{- metric.expires }} |
{{- metric.data_sensitivity|data_sensitivity_numbers }} |
{% endfor %}
{% else %}
This ping contains no metrics.
{% endif %}

{% endfor %}
Data categories are [defined here](https://wiki.mozilla.org/Firefox/Data_Collection).

<!-- AUTOGENERATED BY glean_parser v{{ parser_version }}. DO NOT EDIT. -->
{# The rendered markdown is autogenerated, but this
Jinja2 template is not. Please file bugs! #}
